#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 1A.e - Correction de l'interrogation écrite du 14 novembre 2014
# 
# coût algorithmique, calcul de séries mathématiques

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ## Enoncé 1

# ### Q1
# 
# Le code suivant produit une erreur. Corrigez le programme.

# In[2]:


nbs = [ 1, 5, 4, 7 ]   #
for n in nbs:          #
    s += n             #


# L'objectif de ce petit programme est de calculer la somme des éléments de la liste ``nbs``. L'exception est déclenché la variable ``s`` n'est jamais créé. Il manque l'instruction ``s=0``.

# In[3]:


nbs = [ 1, 5, 4, 7 ]
s = 0
for n in nbs:
    s += n


# In[4]:


s


# ### Q2
# 
# Que vaut ``nbs`` dans le programme suivant :

# In[5]:


def f(x) : return x%2
nbs = { i:f(i) for i in range(0,5) }


# In[6]:


nbs


# ### Q3
# 
# On considère le programme suivant, il affiche ``None``, pourquoi ?

# In[7]:


def ma_fonction(x1,y1,x2,y2):
    d = (x1-x2)**2 +(y1-y2)**2
    print(d)
d = ma_fonction(0,0,1,1)
print(d)


# Le ``2`` correspond au premier ``print(d)``, le ``None`` correspond au second. Pour s'en convaincre, il suffit d'ajouter quelques caractères supplémentaires :

# In[8]:


def ma_fonction(x1,y1,x2,y2):
    d = (x1-x2)**2 +(y1-y2)**2
    print("A",d)
d = ma_fonction(0,0,1,1)
print("B",d)


# Donc la variable ``d`` en dehors de la fonction vaut ``None``, cela veut que le résultat de la fonction ``ma_fonction`` est ``None``. Il peut être ``None`` soit parce que la fonction contient explicitiement l'instruction ``return None`` soit parce qu'aucune instruction ``return`` n'ext exécutée. C'est le cas ici puisqu'il n'y a qu'une instruction ``print``. On remplace ``print`` par ``return``.

# In[9]:


def ma_fonction(x1,y1,x2,y2):
    d = (x1-x2)**2 +(y1-y2)**2
    return d
d = ma_fonction(0,0,1,1)
print(d)


# ### Q4
# 
# Que vaut ``n`` en fonction de ``N`` ?

# In[10]:


n = 0
N = 100
for i in range(0,N):
    for k in range(0,i):
        n += N


# In[11]:


n


# Pour être plus précis, 495000 = $\frac{N^2(N-1)}{2}$.

# ### Q5
# 
# Une des lignes suivantes provoque une erreur, laquelle ?

# In[12]:


a = 3     #
b = "6"   #
a+b       #
a*b       #


# Lorsqu'on multiplie une chaîne de caractères par un entier, cela revient à la répliquer : ``3*"6" = "666"``. L'addition est impossible car on ne peut pas additionner un nombre avec une chaîne de caractères.

# ## Enoncé 2

# ### Q1
# 
# Le code suivant produit une erreur. Proposez une correction.

# In[13]:


nbs = ( 1, 5, 4, 7 )    #
nbs[0] = 0              #


# Le type [tuple](https://docs.python.org/3.4/tutorial/datastructures.html#tuples-and-sequences) sont [immutable](http://fr.wikipedia.org/wiki/Objet_immuable). On ne peut pas le modifier. Mais les listes peuvent l'être.

# In[14]:


nbs = [ 1, 5, 4, 7 ]
nbs[0] = 0


# In[15]:


nbs


# ### Q2
# 
# Que vaut ``c`` ?

# In[16]:


d = {4: 'quatre'}
c = d.get('4', None)


# In[17]:


print(c)


# La méthode [get](https://docs.python.org/3.4/library/stdtypes.html#dict.get) retourne la valeur associée à une clé ou une autre valeur (ici ``None``) si elle ne s'y trouve pas. La raison pour laquelle le résultat est ``None`` ici est que '4' != 4. La clé '4' ne fait pas partie du dictionnaire.

# ### Q3
# 
# Que vaut ``x`` ?

# In[18]:


N = 8
s = 0
while N > 0 :
    for i in range(N):
        s += 1
    N //= 2
x = (s+1)//2


# In[19]:


x


# A chaque passage dans la boucle ``for``, on ajoute ``N`` à ``s``. A chaque passage dans la boucle ``while``, on divise ``N`` par 2. Donc, après la boucle ``while``, $s = N + N/2 + N/4 + N/8 + ...$. On répète cela jusqu'à ce que $N / 2^k$ soit plus grand que 0. Or, les divisions sont entières (symbole ``//``), ``1//2`` vaut 0. La condition devient jusqu'à ce que $N / 2^k <1$. 
# 
# Pour le reste, c'est une suite géométrique. Si on pose $N=2^k$, on calcule donc la somme :
# 
# $$s = 2^k + 2 ^{k-1} + ... + 1 = \sum_{i=1}^{k} 2^i = \frac{2^{k+1}-1}{2-1} = 2^{k+1}-1$$
# 
# Et comme :
# 
# $$x = \frac{s+1}{2} = 2^k = N$$

# ### Q4
# 
# Que vaut ``c`` ?

# In[20]:


l = ['a', 'b', 'c']
c = l[1]


# In[21]:


c


# ### Q5
# 
# Par quoi faut-il remplacer les ``???`` pour avoir l'erreur ci-dessous ?

# In[22]:


def fonction(N):
    li = None             # on évite la variable l pour ne pas la confondre avec 1
    for i in range(N):
        if li is None: 
            li = [ ]
        li.append(i)
    return li
ma_liste = fonction(0)
ma_liste.append(-1)


# Cette erreur se produit car ``ma_liste`` vaut ``None``. Si la fonction ``fonction`` retourne ``None``, c'est que l'instruction ``l = [ ]`` n'est jamais exécutée, donc que la condition ``if l is None`` n'est jamais vérifiée. On ne passe donc jamais dans la boucle ``for`` et ceci arrive si ``N`` est négatif ou nul.

# ## Enoncé 3

# ### Q1
# 
# Que se passe-t-il ?

# In[23]:


l = [ 0, 1,2,3]
for i in range(len(l)):
    print(i)
    del l[i]                #


# L'erreur est due au fait que la boucle parcourt la liste en même temps qu'elle supprime des éléments. Le résultat est souvent une erreur. On vérifie en affichant ``i`` et ``l``.

# In[24]:


l = [ 0, 1,2,3]
for i in range(len(l)):
    print("i=",i,"l=",l)
    del l[i]                  #


# ### Q2
# 
# Que vaut ``a`` ?

# In[25]:


a = 2
for i in range(1,5):
    a += a


# In[26]:


a


# La variable ``a`` double à chaque fois qu'on passe dans la boucle. On y passe **4** fois et on part de ``a=2``. Donc : $2*2*2*2*2=2^5=32$.

# ### Q3
# 
# Que vaut ``y`` ?

# In[27]:


x = 2.67
y = int ( x * 2 ) / 2


# In[28]:


y


# La fonction revient à arrondir au demi inférieur, donc $2.5$.

# ### Q4
# 
# Combien d'étoiles le programme suivant affiche ?

# In[29]:


import random

def moyenne(l):
    s = 0
    for x in l :
        print("*")
        s += x
    return s / len(l)
def variance(l):
    return sum ( [ (x - moyenne(l))**2 for x in l ] ) / len(l)
l = [ random.random() for i in range(0,100) ]
print(variance(l)**0.5)


# C'est un peu long à afficher, modifions le programme pour compter les étoiles plutôt que de les afficher.

# In[30]:


star = 0
def moyenne(l):
    global star
    s = 0
    for x in l :
        star += 1
        s += x
    return s / len(l)
def variance(l):
    return sum ( [ (x - moyenne(l))**2 for x in l ] ) / len(l)
l = [ random.random() for i in range(0,100) ]
print(variance(l)**0.5)
print("star=",star)


# Si $n$ est la longueur de la liste ``l``, le coût de la fonction ``moyenne`` est $O(n)$. Le coût de la fonction ``variance`` est $n$ fois le coût de la fonction ``moyenne``, soit $O(n^2)$. Celle-ci pourrait être beaucoup plus efficace en écrivant :

# In[31]:


star = 0
def moyenne(l):
    global star
    s = 0
    for x in l :
        star += 1
        s += x
    return s / len(l)
def variance(l):
    m = moyenne(l)  # on mémorise le résultat
    return sum ( [ (x - m)**2 for x in l ] ) / len(l)
l = [ random.random() for i in range(0,100) ]
print(variance(l)**0.5)
print("star=",star)


# ### Q5
# 
# Que vaut ``x`` ?

# In[32]:


import random
x = random.randint(0,100)
while x != 50:
    x = random.randint(0,100)


# In[33]:


x


# ``x`` vaut nécessairement 50 puisque c'est la seule valeur qui permette de sortir de la boucle.

# In[34]:





# In[35]: